/*
	ZOONIC RELEASE 2

	(c)1997 by Tinic Urou
*/
#include "zoonic.h"

static sem_id	classlist_sem	= 0;
static Class 	*classlist		= 0;

#ifdef __cplusplus

struct TagItem *FindTagItem(struct TagItem *tags, int TagID )
{	
	if(tags)
	{
		for(;tags->TagID!=TAG_END;tags++)
		{
			if(tags->TagID==TAG_MORE)tags=(struct TagItem *)tags->TagData;
			if(tags->TagID==TagID) return tags;
		}
	}
	return 0;
}

int	GetTagData(struct TagItem *tags, int defval, int TagID )
{
	if(tags)
	{
		for(;tags->TagID!=TAG_END;tags++)
		{
			if(tags->TagID==TAG_MORE)tags=(struct TagItem *)tags->TagData;
			if(tags->TagID==(int)TagID) return (int)tags->TagData;
		}
	}
	return defval;
}

static long 	defaultdispatcher(Class *c, Object *o, ulong MethodID, ulong *data) 
{
	switch(MethodID)
	{
		case	B_METHOD_CONSTRUCT:
		case	B_METHOD_DESTROY:
		case	B_METHOD_SETATTR:
		case	B_METHOD_GETATTR:
				o->SetError(B_NO_ERROR);
				return NULL;
		default:
				o->SetError(B_UNKNOWN_METHOD);
				return NULL;
	}
}

long CreatezOOnic(void)
{
	classlist_sem 				= create_sem(1,"classlist_sem");
	classlist 					= new Class("",0,0,0);
	classlist->_nextclass		= 0;
	classlist->_dispatcher		= defaultdispatcher;
	classlist->_superclass		= 0;
	classlist->_classname		= strdup("root.zclass");
	classlist->_instanceoffset	= 0;
	classlist->_instancesize	= 0;
	classlist->_error			= 0;
	classlist->_image_id		= 0;

	if(classlist&&classlist_sem)return B_NO_ERROR;
	else 						return B_ERROR;
}

void DestroyzOOnic(void)
{
	Class *c=classlist;
	for(;c;)
	{
		classlist=c;c=c->_nextclass;
		
		if(classlist->_superclassname)	free(classlist->_superclassname);
		if(classlist->_classname)		free(classlist->_classname);
		if(classlist->_image_id)		unload_add_on(classlist->_image_id);

		delete classlist;
	}
	delete_sem(classlist_sem);
}

Class *OpenzOOnicClass(char *classname, long minversion)
{
	if(!classname) return classlist;

	if(acquire_sem(classlist_sem)==B_NO_ERROR)
	{
		Class *firstclass=classlist;
		do
		{
			if(firstclass->_classname)
			if(strcmp(firstclass->_classname,classname)==0)
			{
				release_sem(classlist_sem);
				return firstclass;
			}
			firstclass=firstclass->_nextclass;
		}		
		while (firstclass);
		release_sem(classlist_sem);
	}
	
	char filename[MAX_CLASSNAMELENGTH+21]="/boot/system/zclasses/";
	strcat(filename,classname);
	record_ref 	class_ref;
	BFile 		class_file;
	get_ref_for_path(filename,&class_ref);
	class_file.SetRef(class_ref);
	image_id iimage_id = load_add_on(&class_file);

	if(iimage_id != B_ERROR)
	{
		long (*class_version)();
		if(get_image_symbol(iimage_id,"class_version__Fv",						2,&class_version)		!=B_NO_ERROR) return NULL;
		long version=class_version();

		long (*class_superversion)();
		if(get_image_symbol(iimage_id,"class_superversion__Fv",					2,&class_superversion)	!=B_NO_ERROR) return NULL;
		long superversion=class_superversion();

		long (*class_instancesize)();
		if(get_image_symbol(iimage_id,"class_instancesize__Fv",					2,&class_instancesize)	!=B_NO_ERROR) return NULL;
		long instancesize=class_instancesize();

		char *(*class_super)();
		if(get_image_symbol(iimage_id,"class_superclass__Fv",					2,&class_super)			!=B_NO_ERROR) return NULL;
		char *superclassname=class_super();

		long (*dispatcher)(Class *c, Object *o, ulong MethodID, ulong *data);
		if(get_image_symbol(iimage_id,"class_dispatcher__FP5ClassP6ObjectUlPUl",2,&dispatcher)			!=B_NO_ERROR) return NULL;

		if(version>=minversion)
		{
			Class *c = new Class(0,0,0,0);

			c->_instanceoffset	= 0;
			c->_instancesize 	= instancesize;
			c->_version 		= version;
			c->_superclassname	= strdup(superclassname);
			c->_classname 		= strdup(classname);
			c->_image_id 		= iimage_id;
			c->_dispatcher 		= dispatcher;
			c->_nextclass 		= 0;
			c->_superclass 		= OpenzOOnicClass(superclassname,superversion);

			if(c->_superclass)
			{	
				c->_instanceoffset=c->_superclass->_instanceoffset+c->_superclass->_instancesize;

				if(acquire_sem(classlist_sem)==B_NO_ERROR)
				{
					Class *lastclass=classlist;
					for(;lastclass->_nextclass;) lastclass=lastclass->_nextclass;
					lastclass->_nextclass=c;	
					release_sem(classlist_sem);
					
					fprintf(stderr,"%s loaded.\n",classname);
					
					return c;
				}						
			}			
			c->_image_id 		= 0;
//			delete c;
		}
					
		unload_add_on(iimage_id);
	}
	return NULL;
}


Class::Class(char *superclassname, long superminversion, long instancesize, long (*dispatcher)(Class *c, Object *o,ulong MethodID,ulong *data))
{
	if(superclassname==0)	return;
	if(superclassname[0]==0)return;

	_image_id		= 0;
	_superclass		= 0;
	_classname		= 0;
	_error			= B_NO_ERROR; 
	_instancesize	= instancesize;
	_dispatcher		= dispatcher;
	_version		= 0;
	_superclassname = strdup(superclassname);

	if(acquire_sem(classlist_sem)==B_NO_ERROR)
	{
		if(_superclass=OpenzOOnicClass(superclassname,superminversion))
		{				
			if(acquire_sem(classlist_sem)==B_NO_ERROR)
			{
				Class *scanclass;
				for(scanclass=classlist;scanclass->_nextclass;)scanclass=scanclass->_nextclass;	
				scanclass->_nextclass=this;
				_instanceoffset=_superclass->_instanceoffset+_superclass->_instancesize;
				release_sem(classlist_sem);
			}
			else
			{
				_error=B_ERROR;
			}
		}
		else
		{
			_error=B_BADSUPER_CLASS;
		}
		release_sem(classlist_sem);
	}
}

Class::Class(Class *superclass, long instancesize, long (*dispatcher)(Class *c, Object *o, ulong MethodID, ulong *data))
{
	_image_id		= 0;
	_nextclass		= 0;
	_superclass		= 0;
	_classname		= 0;
	_error			= B_NO_ERROR;
	_instancesize	= instancesize;
	_dispatcher		= dispatcher;
	_version		= 0;
	
	if(superclass->_classname)
	{
		_superclassname = strdup(superclass->_classname);
	}	
	else
	{
		_superclassname = 0;
	}
	
	if(acquire_sem(classlist_sem)==B_NO_ERROR)
	{
		Class *scanclass;
		bool found=FALSE;

		for(scanclass=classlist;scanclass->_nextclass;)
		{
			if(superclass==scanclass)
			{
				_superclass=scanclass;
				_instanceoffset=_superclass->_instanceoffset+_superclass->_instancesize;
				found=TRUE;
				break;
			}
			scanclass=scanclass->_nextclass;
		}
		
		if(found)
		{	
			Class *lastclass=classlist;
			for(;lastclass->_nextclass;)lastclass=lastclass->_nextclass;	
			lastclass->_nextclass=this;
		}
		else
		{
			_error=B_UNKNOWN_CLASS;
		}		

		release_sem(classlist_sem);
	}
}

Class::~Class()
{
	if(acquire_sem(classlist_sem)==B_NO_ERROR)
	{
		Class *scanclass=classlist,*prevclass=0;
		
		bool found=FALSE;

		for(;scanclass->_nextclass;)
		{
			if(this==scanclass)
			{
				found=TRUE;
				break;
			}
			prevclass=scanclass;
			scanclass=scanclass->_nextclass;
		}
		
		if(prevclass)
		{
			prevclass->_nextclass=scanclass->_nextclass;
		}
		else
		{
			classlist=scanclass->_nextclass;
		}
				
		release_sem(classlist_sem);
	}
}

void Class::SetError(long errorcode)
{
	_error=errorcode;
};

long Class::Error()
{
	return _error;
};

Object::Object(Class *baseclass, ...)
{
	_instance=0;
	_baseclass=0;

	if(_baseclass=baseclass)
	{
		_instance = (void *)(new uchar [_baseclass->_instanceoffset+_baseclass->_instancesize+4]);
		
		va_list	argptr;
		va_start (argptr,baseclass);
		_baseclass->_dispatcher(_baseclass,this,B_METHOD_CONSTRUCT,(ulong *)argptr); 
		va_end (argptr);
		
		_error=B_NO_ERROR;
		return;
	}

	_error=B_ERROR;
};

Object::Object(char *baseclassname, long minversion ...)
{
	_instance=0;
	_baseclass=0;
	if(baseclassname)
	{
		if(_baseclass = OpenzOOnicClass(baseclassname, minversion))
		{	
			_instance = (void *)(new uchar [_baseclass->_instanceoffset+_baseclass->_instancesize+4]);

			va_list argptr;
			va_start (argptr,minversion);
			_baseclass->_dispatcher(_baseclass,this,B_METHOD_CONSTRUCT,(ulong *)argptr);
			va_end (argptr);

			_error=B_NO_ERROR;
			return;
		}
	}
	_error=B_ERROR;
};

Object::~Object()
{
	if(_baseclass)
	{
		_baseclass->_dispatcher(_baseclass,this,B_METHOD_DESTROY,0);
	}
	if(_instance)
	{
		delete [] ((uchar *)_instance);
	}
}
#endif
