/*
	Skeleton for Player class
*/
#include <InterfaceKit.h>
#include "Player.h"

#define PLAYER_NAME "Your player name"

// This is the instance of our class. Since player objects are only 
// created once you can also use global variables... For better OO
// look use preferably this instance.

struct instance_struct
{
	BView		*prefs_view;
	BMessage 	*prefs;
	bool		loop;
	int			volume;
	int 		pos;
	int			track;
};

// Here is were all interesting happens.
// Normally DualPlayer will call this dispatcher and you should
// process the requested method. If you want to call a method on
// your own use: o->DoMethod(METHOD,arg1,arg2,...);

long class_dispatcher(Class *c, Object *o, ulong MethodID, ulong *data)
{
	instance_struct *inst=(instance_struct *)o->Instance(c);	

	o->SetError(B_NO_ERROR);

	switch(MethodID)
	{

		// The constructor of the zOOnic class. Its like the
		// C++ constructor

		case	B_METHOD_CONSTRUCT:	// void (void)
				{					
					inst->prefs_view = new BView(BRect(0,0,0,0),"",NULL,NULL);
					inst->prefs = NULL;
					return 0;
				}
		// The destructor

		case	B_METHOD_DESTROY:	// void (void)
				{
					return 0;
				}				

		// Load the a file, start and prepare that the 
		// stream method is called. return a TRUE value
		// or a handle on return when loading succeded.
		
		// TIP: Most available players do not offer a callback
		// playing. Preferably spawn a thread here and use semaphores
		// to sycronise to the audio stream.

		// This method has an argument. Arguments can be found
		// in the 'char *data' pointer. Here the pointer to
		// the filename/path is in data[0].
				
		case	P_LOAD:				// bool handle (char *path)
				{			
					char *path=(char *)data[0];

					return FALSE; // No success
				}
		
		// Free the file
									
		case	P_FREE:				// void (int handle)
				{
					int handle=(int)data[0];
					
					return 0;
				}

		// Not used yet

		case	P_RESUME:			// void (int handle)
				{
					int handle=(int)data[0];
					return 0;
				}

		// Not used yet

		case	P_PAUSE:			// void (int handle)
				{
					int handle=(int)data[0];
					return 0;
				}

		// Here you have the possibility to setup
		// the AudioSubscriber
					
		case	P_SETUPSTREAM:		// void (BAudioSubscriber *auds)
				{
					BAudioSubscriber *auds=(BAudioSubscriber *)data[0];

//					auds->SetDACSampleInfo(2, 2, B_BIG_ENDIAN, B_LINEAR_SAMPLES);
//					auds->SetSamplingRate(44100);					

					return 0;
				}

		// The stream function. Like the AudioSubscriber 
		// stream function.
				
		case	P_STREAM:			// bool (void *arg, char *buffer, int size)
				{
					void *arg		=(void *)data[0];
					char *buffer	=(char *)data[1];
					int size		=(int)data[2];
					
					return 0; // Do not forget to change this one otherwise you wont hear anything ;-)
				}

		// This method should be the only way to access the
		// instance and to set attributes. The data pointer
		// is here a pointer to a TagList.

		// Its not a problem to perform actions here...
				
		case	P_SETATTR:			//  void (ulong AttributeID, ulong data)
				{
					ulong ATTR =(ulong)data[0];
					ulong DATA =(ulong)data[1];

					switch(ATTR)
					{
						case	P_ATTR_TRACK:				// Forward or backward to a track
								inst->track=(int)DATA;
								break;
						case	P_ATTR_PREFS:				// Setup the pointer to the preferences BMessage
								inst->prefs=(BMessage *)DATA;
								break;
						case	P_ATTR_BOOST:				// Set volume 0..128, special boost from 128..255
								inst->volume=(int)DATA;
								break;
						case	P_ATTR_MODPOS:				// Set position in the module
								inst->pos=(int)DATA;
								break;
						case	P_ATTR_LOOP:				// Loop the file?
								inst->loop=(bool)DATA;
								break;
					}
					o->DoSuperMethod(c,P_SETATTR,data);
					return 0;
				}

		// This method should be the only way to access the
		// instance and to get attributes.
				
		case	P_GETATTR:			//  void (ulong AttributeID, ulong data)
				{
					ulong ATTR =(ulong)data[0];
					ulong DATA =(ulong)data[1];

					switch(ATTR)
					{
						case	P_ATTR_STREAMARG:			// Argument which should be forwarded to the stream function or 0
								*((int *)DATA)=0;
								return 0;
						case	P_ATTR_ID:					// ID String which is displayed in the ListView, this must be valid!
								*((char **)DATA)=PLAYER_NAME;
								return 0;
						case	P_ATTR_VIEW:				// Get the pointer to the Preferences BView, this must be valid!
								*((BView **)DATA)=inst->prefs_view;
								return 0;
						case	P_ATTR_MODLENGTH:			// Number of position in the file or 0
								*((int *)DATA)=0;
								return 0;
						case	P_ATTR_MODPOS:				// Position in the file or 0
								*((int *)DATA)=0;
								return 0;
						case	P_ATTR_TRACK:				// return 0 when there are no tracks.
								*((int *)DATA)=0;
								return 0;
						case	P_ATTR_TRACKNUM:			// return 0 when there are no tracks.
								*((int *)DATA)=0;
								return 0;
						default:
								return o->DoSuperMethod(c,P_GETATTR,data);
					}
				}

		// Draw stats in the main window
		// Only draw to the supplied BView.
				
		case	P_DRAWSTATUS:		// void (BView *view)
				{
					BView *view=(BView *)data[0];
					
					return 0;
				}
		
		// Create module info. Set *lines to the number of used lines. *info
		// should be an array containing the strings. *info and the
		// strings must be allocated on your own.
				
		case	P_CREATEINFO: 		// void (char ***info, int *lines)
				{
					char ***info=(char ***)data[0];
					int *lines=(int *)data[1];

					*info=0;	// Safety settings
					*lines=0;
																		
					return 0;
				}

		// Free the module info

		case	P_FREEINFO: 		// void (char ***info, int *lines)
				{
					char ***info=(char ***)data[0];
					int *lines=(int *)data[1];
					
					return 0;
				}

		// Redirect any unknown method to the superclass
		default:
				return o->DoSuperMethod(c,MethodID,data);
	}
}

char *class_superclass()	// Superclassname for ALL players
{
	return "dualplayer/player/baseplayer.zclass";
};

long class_instancesize()
{
	return sizeof(struct instance_struct);
};

long class_superversion()	// minimum version needed of the superclass
{
	return 2;
};

long class_version()		// This class version
{
	return 2;
};
